#if ! defined( _CPF_H_ )
#define _CPF_H_

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "Element.h"
#include "RotAngles.h"
#include "eop.h"
#include "Offsets.h"

template <class T>
inline long sign( T a, T b, T c, T x, T y, T z  )
{
    if( a*x >= T(0) && b*y >= T(0) && c*z >= T(0) )return -1;
    if( a*x <= T(0) && b*y <= T(0) && c*z <= T(0) )return 1;
    yourTHROW("Incompatible data.");
}

//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// The Cpf class is the set of unique data associated with one CPF or 
// Consolidated Prediction Format prediction set.
//
// Remarks: This set of data consist of the synchronous CPF data records,
// i.e. those aligned with the position record epoch. Asynchronous data
// such as target offset, rotation angles and eop data have their own classes.
//-----------------------------------------------------------------------------
class Cpf : public Element
{
public :

    //  Constructor
    Cpf();

    //  Destructor
    virtual ~Cpf();

    // CPF data records (version 1)
    enum RecordType
    { 
        unknownRecordType = 0,
        headerBasic1,
        headerBasic2,
        headerAccuracy,
        headerTransponder,
        headerCoM,
        lastHeader,
        recordPosition,
        recordVelocity,
        recordCorrection,
        recordTransponder,
        recordOffset,
        recordAngle,
        recordEOP,
        lastRecord
    };

    enum ReferenceFrame
    {
        unknownRefFrame = 0,
        itrf,                   // geocentric true body fixed ITRF (default)
        tod,                    // geocentric space fixed (i.e. Inertial) (True of Date)
        mod                     // geocentric space fixed (Mean of Date J2000)
    };

    enum TargetType
    {
        unknownTargetType = 0,
        passiveSatellite,       // passive artificial satellite
        passiveLunar,           // passive lunar reflector
        synchTransponder,       // synchronous transponder
        asynchTransponder       // asynchronous transponder
    };

    // Structure to support transponder data, when provided
    struct Transponder
    {
        Transponder()
        {
            pulseFrequency = transmitDelay = utcOffset =0.0;
            oscillatorDrift = oscillatorRelCor = 0.0;
        }
        double pulseFrequency;    // Hz
        double transmitDelay;     // microsecs
        double utcOffset;         // microsecs
        double oscillatorDrift;   // parts in 10^15
        double oscillatorRelCor;  // meters/second
    };

    // Structure to support accuracy/runoff data, when provided
    struct Accuracy
    {
        Accuracy()
        {
            for( long i = 0; i < 3; i++ )
                alongTrack[i]  = acrossTrack[i] = radial[i]= 0.0;
        }
        double alongTrack[3];   // meters
        double acrossTrack[3];  // meters
        double radial[3];       // meters
    };


    // Method to determine the type of record.
    static RecordType recordType( const std::string& s);

    //===============
    // Set methods
    //===============

    //  Method to set the Satellite Identifier code
    void sicIs(std::string sic){m_sic = sic;};

    //  Method to set the Norad Catalog number
    void noradIs(std::string num){m_norad = num;};

    //  Method to set the COSPAR identifier.
    void cosparIs(std::string cospar){m_cospar = cospar;};

    //  Method to set the ephemeris sequence number.
    void sequenceIs(long seq) {m_sequence = seq;};    

    //  Method to set the x y and z-position in meters
    void xPositionIs  (double x){m_xPositionTx = x;};   
    void yPositionIs  (double y){m_yPositionTx = y;};    
    void zPositionIs  (double z){m_zPositionTx = z;}; 
    void xPositionRxIs(double x){m_xPositionRx = x;};   
    void yPositionRxIs(double y){m_yPositionRx = y;};    
    void zPositionRxIs(double z){m_zPositionRx = z;};    

    //  Method to set the x y and z-velocities in meters/sec
    void xVelocityIs  (double u){m_xVelocityTx = u;};   
    void yVelocityIs  (double v){m_yVelocityTx = v;};  
    void zVelocityIs  (double w){m_zVelocityTx = w;};  
    void xVelocityRxIs(double u){m_xVelocityRx = u;};   
    void yVelocityRxIs(double v){m_yVelocityRx = v;};  
    void zVelocityRxIs(double w){m_zVelocityRx = w;};  

    //  Methods to set the stellar aberration corrections in meters.
    void xAberrationCorTxIs( double cor ){ m_xAberrationTx = cor;};
    void yAberrationCorTxIs( double cor ){ m_yAberrationTx = cor;};
    void zAberrationCorTxIs( double cor ){ m_zAberrationTx = cor;};
    void xAberrationCorRxIs( double cor ){ m_xAberrationRx = cor;};
    void yAberrationCorRxIs( double cor ){ m_yAberrationRx = cor;};
    void zAberrationCorRxIs( double cor ){ m_zAberrationRx = cor;};

    //  Methods to set the relativistic range corrections in nsecs.
    void rangeRelCorTxIs( double cor ){ m_relCorrTx = cor;};
    void rangeRelCorRxIs( double cor ){ m_relCorrRx = cor;};

    //  Method to set a pointer to a transponder structure.
    void transponderIs( Transponder transponder ){ m_transponder = transponder;};

    //  Method to set the reference frame.
    void refFrameIs( ReferenceFrame frame ){ m_referenceFrame = frame;};

    //  Method to set the IRV compatability.
    void irvCompatibleIs( bool compat){ m_irvCompatible = compat;};

    //  Method to set the rotation angle type.
    void rotationAngleTypeIs( RotationAngles::Type type ){ m_angleType = type;};

    //  Method to set the target cpf type.
    void targetTypeIs( TargetType type){ m_targetType = type;};

    //===============
    // Access methods
    //===============
    //  Method to access the Object Identifier code
    std::string sic() const {return m_sic;};

    //  Method to access the Norad catalog number
    std::string norad() const {return m_norad;};

    //  Method to access the International Cospar Identifier
    std::string cospar() const {return m_cospar;};  

    //  Method to access the ephemeris sequence number.
    long sequence() const {return m_sequence;};    

    //  Method to access the outbound x, y or z position in meters and Mm.
    double xPosition()  const{return m_xPositionTx;};   
    double yPosition()  const{return m_yPositionTx;};    
    double zPosition()  const{return m_zPositionTx;};    
    double xPositionMm()const{return m_xPositionTx * 1.0e-6;};   
    double yPositionMm()const{return m_yPositionTx * 1.0e-6;};    
    double zPositionMm()const{return m_zPositionTx * 1.0e-6;};    

    //  Method to access the inbound x, y or z position in meters and Mm.
    double xPositionRx()  const{return m_xPositionRx;};   
    double yPositionRx()  const{return m_yPositionRx;};    
    double zPositionRx()  const{return m_zPositionRx;};    
    double xPositionRxMm()const{return m_xPositionRx * 1.0e-6;};   
    double yPositionRxMm()const{return m_yPositionRx * 1.0e-6;};    
    double zPositionRxMm()const{return m_zPositionRx * 1.0e-6;};    

    //  Method to access the outbound x, y or z velocities in meters/sec and Mm/day.
    double xVelocity()        const{return m_xVelocityTx;};   
    double yVelocity()        const{return m_yVelocityTx;};  
    double zVelocity()        const{return m_zVelocityTx;};  
    double xVelocityMmPerDay()const{return m_xVelocityTx * 1.0e-6 * SECS_PER_DAY;};   
    double yVelocityMmPerDay()const{return m_yVelocityTx * 1.0e-6 * SECS_PER_DAY;};  
    double zVelocityMmPerDay()const{return m_zVelocityTx * 1.0e-6 * SECS_PER_DAY;};  

    //  Method to access the inbound x, y or z velocities in meters/sec and Mm/day.
    double xVelocityRx()        const{return m_xVelocityRx;};   
    double yVelocityRx()        const{return m_yVelocityRx;};  
    double zVelocityRx()        const{return m_zVelocityRx;};  
    double xVelocityRxMmPerDay()const{return m_xVelocityRx * 1.0e-6 * SECS_PER_DAY;};   
    double yVelocityRxMmPerDay()const{return m_yVelocityRx * 1.0e-6 * SECS_PER_DAY;};  
    double zVelocityRxMmPerDay()const{return m_zVelocityRx * 1.0e-6 * SECS_PER_DAY;};  

    //  Method to access the outbound x, y or z stella aberration in meters.
    double xAberrationCorTx() const {return m_xAberrationTx; };
    double yAberrationCorTx() const {return m_yAberrationTx; };
    double zAberrationCorTx() const {return m_zAberrationTx; };

    //  Method to access the inbound x, y or z stella aberration in meters.
    double xAberrationCorRx() const {return m_xAberrationRx; };
    double yAberrationCorRx() const {return m_yAberrationRx; };
    double zAberrationCorRx() const {return m_zAberrationRx; };

    //  Method to access the outbound or inbound range relativistic correction in nsec.
    double rangeRelCorRx() const {return m_relCorrRx; };
    double rangeRelCorTx() const {return m_relCorrTx; };

    //  Method to access transponder object and data if available.
    Transponder transponder() const { return m_transponder;};

    //  Method to access the reference frame.
    ReferenceFrame refFrame() const {return m_referenceFrame;}; 

    //  Method to access the IRV compatability.
    bool irvCompatible() const {return m_irvCompatible;};

    //  Method to access the IRV compatability.
    bool lightTimeCorrected() const {return m_lightTimeCorrected;};

    //  Method to access the bidirectional flag. Returns
    // true if independent outbound and inbound position data are available.
    bool biDirectional() const {return m_biDirectionalPos;};
    bool biDirectionalVel() const {return m_biDirectionalVel;};
    bool biDirectionalCor() const {return m_biDirectionalCor;};

    //  Method to access the bidirectional flag. Returns
    // true if independent outbound and inbound position data are available.
    bool valid() const {return m_complete;};

    //  Method to access the rotation angle type.
    RotationAngles::Type rotationAngleType() const {return m_angleType;}; 

    //  Method to access the target cpf type.
    TargetType targetType() const {return m_targetType;}; 

    //  Method to access the target centre of mass correction (m).
    double centreOfMassCorr() const {return (m_comApplied ? 0 : m_comCorr);}; 


protected :

    friend class PredictorCPF;

    struct Header
    {
        Header()
        {
            complete = false;
            comApplied = irvCompatible = false;
            interval = sequence = 0;
            targetType = Cpf::TargetType::unknownTargetType;
            referenceFrame= Cpf::ReferenceFrame::unknownRefFrame;
            angleType = RotationAngles::Type::unknownAngleType;
            nh1 = nh2 = nh3 = nh4 = nh5 = nh9 = identifier = 0;
            centreOfMassCorr = 0.0;
        }
        bool complete;
        std::string source;
        std::string target;
        std::string notes;
        std::string cospar;
        std::string norad;
        std::string sic;
        YourTimeDateClass created;
        YourTimeDateClass startDate;
        YourTimeDateClass endDate;
        long sequence;
        long interval;
        bool irvCompatible;
        bool comApplied;
        Cpf::TargetType targetType;
        Cpf::ReferenceFrame referenceFrame;
        RotationAngles::Type angleType;
        Accuracy runoff;
        Transponder transponder;
        double centreOfMassCorr;
        long nh1, nh2, nh3, nh4, nh5, nh9;
        long identifier;
    };

    // Methods to parse each CPF header and data record.
    static void parseBasicHeader1(       const std::string& line, Header* header );
    static void parseBasicHeader2(       const std::string& line, Header* header );
    static void parseAccuracyHeader(     const std::string& line, Header* header );
    static void parseTransponderHeader(  const std::string& line, Header* header );
    static void parseCentreOfMassHeader( const std::string& line, Header* header );
    static void parseLastHeader(         const std::string& line, Header* header );
    void parsePositionRecord(    const std::string& line, const Header& header );
    void parseTransponderRecord( const std::string& line, const Header& header ); 
    void parseVelocityRecord(    const std::string& line, const Header& header );
    void parseCorrectionRecord(  const std::string& line );
    void parseLastRecord(        const std::string& line );

    // Method to add a pointer to each CPF record to the other data objects.
    void addHeaderData( const Header& header );

    YourTimeDateClass        m_epochRx;
    std::string               m_cospar;
    std::string               m_norad;
    std::string               m_sic;
    long                      m_sequence;
    TargetType                m_targetType;
    ReferenceFrame            m_referenceFrame;
    bool                      m_irvCompatible;
    bool                      m_comApplied;
    RotationAngles::Type m_angleType;

    double   m_xPositionTx;      // out x-position in meters
    double   m_yPositionTx;      // out y-position in meters
    double   m_zPositionTx;      // out z-position in meters
    double   m_xVelocityTx;      // out x-velocity in meters/sec
    double   m_yVelocityTx;      // out y-velocity in meters/sec
    double   m_zVelocityTx;      // out z-velocity in meters/sec
    double   m_xPositionRx;      // back x-position in meters
    double   m_yPositionRx;      // back y-position in meters
    double   m_zPositionRx;      // back z-position in meters
    double   m_xVelocityRx;      // back x-velocity in meters/sec
    double   m_yVelocityRx;      // back y-velocity in meters/sec
    double   m_zVelocityRx;      // back z-velocity in meters/sec
    double   m_xAberrationTx;    // Out X stellar aberration correction in meters
    double   m_yAberrationTx;    // Out Y stellar aberration correction in meters
    double   m_zAberrationTx;    // Out Z stellar aberration correction in meters
    double   m_xAberrationRx;    // Back X stellar aberration correction in meters
    double   m_yAberrationRx;    // Back Y stellar aberration correction in meters
    double   m_zAberrationRx;    // Back Z stellar aberration correction in meters
    double   m_relCorrTx;        // Out Relativistic range correction in nsec (positive number)
    double   m_relCorrRx;        // Back Relativistic range correction in nsec (positive number)
    double   m_comCorr;          // Centre of mass correction, positive meters.
    Transponder m_transponder;

    bool m_lightTimeCorrected;
    bool m_biDirectionalPos;
    bool m_biDirectionalVel;
    bool m_biDirectionalCor;
    bool m_complete;
    long m_nlr;

};  // class Cpf

#endif

