//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include "Offsets.h"
#include "yourException.h"
#include <math.h>

/******************************************************************************
 Constructor 
******************************************************************************/
TargetOffsets::TargetOffsets()
{
    m_biDirectional = false;
    m_next = 0;
}
 
/******************************************************************************
 Destructor
******************************************************************************/
TargetOffsets::~TargetOffsets(  )
{
}

/******************************************************************************
 This subroutine uses linear interpolation to obtain target offset values
 at a given epoch.

 The function returns FALSE if the caller supplies an epoch outside the 
 range of available data. The function does not extrapolate data.
******************************************************************************/
bool TargetOffsets::getOut(YourTimeDateClass epoch,
                           double& xOut, double& yOut, double& zOut )const
{
    static bool alertGiven = false;
    xOut = yOut = zOut = 0.0;

    if( m_aOffsets.empty() )
        return false;

    std::vector<TargetOffsets::Record>::const_iterator i = m_aOffsets.begin();
    if( m_aOffsets.size() == 1 )
    {
        xOut = i->xOut;
        yOut = i->yOut;
        zOut = i->zOut;
        return true;
    }
    else
    {
        for( ; i < m_aOffsets.end(); i++ )
        {
            double d1 = YourTimeIntervalClass::days(        epoch - i->epoch );
            double d2 = YourTimeIntervalClass::days( (i+1)->epoch -    epoch );
            double d3 = YourTimeIntervalClass::days( (i+1)->epoch - i->epoch );
            if( fabs(d1) <= fabs(d3) && fabs(d2) <= fabs(d3) )
            {
                double f = 0.0;
                if( fabs(d3) > 0.0 ) 
                    f = d1 / d3;
                xOut =  ( (i+1)->xOut - i->xOut ) * f + i->xOut;
                yOut =  ( (i+1)->yOut - i->yOut ) * f + i->yOut;
                zOut =  ( (i+1)->zOut - i->zOut ) * f + i->zOut;
                return true;
            }
        }
    }

    return false;
}

/******************************************************************************
 This subroutine uses linear interpolation to obtain target offset values
 at a given epoch.

 The function returns FALSE if the caller supplies an epoch outside the 
 range of available data. The function does not extrapolate data.
******************************************************************************/
bool TargetOffsets::getIn(YourTimeDateClass epoch,
                          double& xIn,  double& yIn,  double& zIn )const
{
    xIn  = yIn  = zIn  = 0.0;

    if( m_aOffsets.empty() )
        return false;

    std::vector<TargetOffsets::Record>::const_iterator i = m_aOffsets.begin();
    if( m_aOffsets.size() == 1 )
    {
        xIn  = i->xIn;
        yIn  = i->yIn;
        zIn  = i->zIn;
        return true;
    }
    else
    {
        for( ; i < m_aOffsets.end(); i++ )
        {
            double d1 = YourTimeIntervalClass::days(        epoch   - i->epochRx );
            double d2 = YourTimeIntervalClass::days( (i+1)->epochRx -    epoch   );
            double d3 = YourTimeIntervalClass::days( (i+1)->epochRx - i->epochRx );
            if( fabs(d1) <= fabs(d3) && fabs(d2) <= fabs(d3) )
            {
                double f = 0.0;
                if( fabs(d3) > 0.0 ) 
                    f = d1 / d3;
                xIn  =  ( (i+1)->xIn  - i->xIn  ) * f + i->xIn;
                yIn  =  ( (i+1)->yIn  - i->yIn  ) * f + i->yIn;
                zIn  =  ( (i+1)->zIn  - i->zIn  ) * f + i->zIn;
                return true;
            }
        }
    }

    return false;
}

/******************************************************************************
CPF record type 5 Offset from center of main body
(Surface features and satellites)

 1-2 A2 Record Type (= "50")
 I1 Direction flag (bounce=0; transmit = 1; receive = 2)
I5  Modified Julian Date (MJD)
f13.6 Seconds of Day (UTC)
a10 Name of target (no spaces in middle)
f17.3 X position offset in meters
f17.3 Y position offset in meters
f17.3 Z position offset in meters
******************************************************************************/
void TargetOffsets::addCpfRecord( const std::string& line )
{
    std::istringstream s;
    s.str( line );

    bool ok = true;
    std::string recType = convert<std::string>( s, ok );
    if( recType != "50" )
        yourTHROW( "Error reading CPF record 50.");

    long direction   = convert<long>( s, ok );       
    double mjd       = convert<double>( s, ok );     
    double secs      = convert<double>( s, ok );     
    std::string name = convert<std::string>( s, ok );
    double xoff      = convert<double>( s, ok );     
    double yoff      = convert<double>( s, ok );     
    double zoff      = convert<double>( s, ok );     
    if( !ok )
        yourTHROW( "Error reading CPF record 50.");

    if( !m_name.empty() && m_name != name)
        yourTHROW( "Error reading CPF record 50. Inconsistent name.");
    m_name = name;
    YourTimeDateClass epoch = YourTimeIntervalClass::modJulianDate( mjd, 0 );
    epoch += YourTimeIntervalClass::seconds(secs, 15);

    if( direction == 0 || direction == 1 )
    {
        Record offset;
        offset.epoch = epoch;
        offset.xOut = xoff;
        offset.yOut = yoff;
        offset.zOut = zoff;
        offset.epochRx = epoch;
        offset.xIn = -xoff;
        offset.yIn = -yoff;
        offset.zIn = -zoff;
        m_aOffsets.push_back( offset );
    }

    if( direction == 0 )
    {
        if( m_next != 0 )
            yourTHROW( "Directional offset records out of sequence.");
    }
    else if( direction == 1 )
    {
        if( m_biDirectional && m_next != 1 )
            yourTHROW( "Directional offset records out of sequence.");
        m_next = 2;
    }
    else if( direction == 2 )
    {
        if( m_next != 2 )
            yourTHROW( "Directional offset records out of sequence.");
        m_biDirectional = true;

        TargetOffsets::Record& prev = m_aOffsets.back();
        long sgn = 1;
        if( prev.xOut * xoff > 0.0 )
            sgn = -1;
        prev.epochRx = epoch;
        prev.xIn = sgn*xoff;
        prev.yIn = sgn*yoff;
        prev.zIn = sgn*zoff;
        m_next = 1;
    }
    return;
}
