#pragma warning( disable : 4786 )

#if ! defined( _TRACK_H_ )
#define _TRACK_H_

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <string>
#include <vector>
#include "Sitehandle.h"
#include "yourTimeClass.h"
#include "yourMatrixClass.h"


//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
 
// The Track class is a base class for providing data and methods
// of managing track data for a particular target (of a particular type)
// over a defined time period. It is a container for a series of target 
// ephemeris (position and velocity) and other data in a given reference frame,
// this being the true earth (body) fixed or ITRF with x axis refernced to
// Greenwich prime meridian, and z in axis of earth rotation. 
//
// Once a site is defined, by creating a SiteHandle object, track data in the
// form of topocentric pointing angles (azimuth, elevation, range) can be 
// accessed for any given epoch within the defined time period.
// Derived classes allow the creation of track objects having different internal
// data structures and coordinate transformation methods.
//
// The data stored in these classes will be generated by a "Predictions
// Generator" for satellites and luner targets. This class may be extened to include
// other types of targets which may not require prediction generators.
//
// For a given epoch and site location, the common access methods provides;
// 1. azimuth
// 2. elevation
// 3. range
// 4. proximity to the sun
// 5. proximity to the moon
// 6. magnitude
// 7. azimuth velocity
// 8. elevation velocity
// 9. range velocity
// 10. speed on the azimuth-elevation plane
// 11. timebias
// 12. offset biases
//
// Refer to descended classes for more specific information.
//
// Example: 
//   // Construct a track object, trk ( of appropriate derived class TrackClass )
//   TrackClass trk( start, stop, stepsize );
//
//   // If necessary, create the required type of "predictor" to generate prediction data.
//   PredictorClass predictor( trk, "targetName", "directory");
//   loop
//       trk.appendNode( time, .... );
//   end loop
//
//   // Once trk contains basic predictions data, it can then used to provide 
//   // predictions at any time within the prediction period for the required sites.
//   // First construct a "SiteHandle" for a site.
//   SiteHandle s1( latitude, longitude, height ... );
//
//   // Calculate position for the given site at a given time or epoch, t. e.g.
//   trk.calculatePosition( s1, t);
//
//   // Now access topocentric (and other) data
//   double az = s1.azimuthDegree();
//   double el = s1.elevationDegree();
//   double r  = s1.rangeMm();
//    etc.
//-----------------------------------------------------------------------------
class Track
{
protected:

    //  Default constructor.
    Track(){};

    //  Constructor
    Track( YourTimeDateClass     startPredictions,
           YourTimeDateClass     stopPredictions,
           YourTimeIntervalClass stepSize);

public:

    //  Destructor
    virtual ~Track();
    
    //  Enumeration of defined track types.
    enum Type
    {
        undefined = 0,   // unknown track type
        orbital,         // orbital track derived from earth centred, earth fixed X, Y, Z
        lunar,           // track of a lunar target for LLR
        transponder,     // track of a transponder 
    };

    // This method must be called to generate current target track/ephemeris/predictions.
    // For a given time and observation site, this method generates interpolated values using
    // stored track data, and then calculates the required range, pointing angles and other 
    // current data which can be accessed using the available access methods.
    // Arguments: epoch -  time at whcuh data is required
    //             site -  information on a given obervation site
    // Return Value: True if data successfully found and processed. False otherwise.
    // Note:  If there are no prediction data available for the requested time and date,
    // false is returned. Extrapolation is not supported. 
    bool calculatePosition(SiteHandle& site, YourTimeDateClass epoch);

    // This method must be called to obtain current target track/ephemeris/predictions.
    // For a given node index and observation site, this method extracts values of 
    // stored track data, and then calculates the required range, pointing angles and other 
    // current data which can be accessed using the available access methods.
    // Arguments: index -  index of node, 0 to number of nodes stored in the track.
    //             site -  information on a given obervation site
    // Return Value: True if data successfully found and processed. False otherwise.
    // Note:  If the requested index is outside the range of available predictions,
    // false is returned.  
    bool calculatePosition(SiteHandle& site, long index);
    
    // Method to append a prediction node which includes
    // position, velocity and stella aberration correction data. 
    // Arguments: posVecCor -  a matrix containing the following data;
    // geocentric ITRF x,y,z positions at next time interval (Megameters)
    // geocentric ITRF u,v,w velocities at next time interval (Megameter/day)
    // x,y,z stellar abberration corrections at next time interval (m).
    // polar motion data, xp, yp (arcsecs) and UT1-UTC at next time interval.
    // biDirectional flag to indicate if independent inbound data is available. 
    // newSegment - flag indicating that this point is first in a new segment, i.e.
    //              predictions using a new orbital element.
    virtual void appendNode( YourTimeDateClass epoch,
                             bool newSegment, bool biDirectional, bool lightTimeCorrected,
                             const YourMatrix<double>& posVelCor, double xPolar,  double yPolar,
                             YourTimeIntervalClass ut1_utc ){return;};
    //------------------------------
    // Set parameters
    //------------------------------
    //  Method to set the start and end epochs of the period for which
    // prediction data are currently required. These start and stop epochs can be redefined
    // once the track object has been created. This period may contain one or passes, but the 
    // start and stop interval should be within the period of available prediction
    // points as specified in the constructor. 
    // Arguments: epoch - time
    // Note - if the start epoch is earlier than available predictions, or the stop epoch
    // is later than available predictions then an exception is raised. If necessary, 
    // call predictionStartEpoch() to check earliest available prediction actually generated.
    void setStartEpoch(YourTimeDateClass epoch);
    void setStopEpoch(YourTimeDateClass epoch);

    //  Method to reset the node interval (ie step size). This must only
    // be called before the track object is populated with prediction data.
    // This method should only be used in specific situations such as when the predictor can only
    // provide node data at a predetermined interval which is unknown to the client.
    // This value will overwrite that provided by the constructor.
    void setStepSize( YourTimeIntervalClass interval){ m_stepSize = interval;};
 
    //  By default, the best available quality predictions are
    // created, i.e. that the correct number of nodes are generated and the highest
    // fidelity models are used for calculating predictions. There are 
    // situations where this may not be desirable. Courser predictions may be sufficient
    // giving a signifcant reduction in calculation time and memory usage. Call this
    // method to reduce prediction fidelity. 
    void bestPredictionsNotRequired(){m_bestPrecision = false;};

    //  Method to set the interpolator width, i.e. the number of points
    //  used by the interpolator which effects the level of smoothing.
    // Arguments: number - number of points used by the interpolator
    // Note: In general it should not be necessary to modify the default value
    // which is set to 8.
    void setInterpolatorWidth(long number);

    //---------------
    // Access Methods
    //---------------
    //  Method to access the start and stop times of the prediction period.
    Type trackType() const {return m_trackType;};

    //  Method to access the start and stop times of the requested track period.
    YourTimeDateClass periodStart() const { return m_periodStart; }
    YourTimeDateClass periodStop() const { return m_periodStop; }

    //  Method to access the start and stop times of the prediction period.
    YourTimeDateClass predictionStartEpoch() const { return m_startPredicts; }
    YourTimeDateClass predictionStopEpoch() const { return m_stopPredicts; }

    //  Method to access the end of the required prediction period.
    // Extended period can be used if internal orbit predictions are required to be
    // generated beyond the nominal stop epoch.  
    YourTimeDateClass startEpoch() const {return m_startEpoch;}
    YourTimeDateClass stopEpoch() const {return m_stopEpoch;}

    //  Method to access the time interval between nodes. 
    // Note: The time interval between nodes is constant.
    YourTimeIntervalClass nodeInterval() const {return m_stepSize;}

    //  Method to access the number of nodes generated. 
    long numberOfnodes() const {return m_numOfNodes;}

    //  Method to access the defined precision level used
    // when generating predictions.
    bool bestPrecision() const { return m_bestPrecision;}

    //  Method to access the interpolator width, i.e. the
    // number of points used by the interpolator which effects the level of
    // smoothing.
    long interpolatorWidth() const {return m_numInterpMax;}

protected:

    //  This function generates Lagragian interpolation coefficients based on
    // required time and time of surrounding nodes.
    // Arguments: number - Number of nodes to be used in generating the Lagrangian coefficients.
    //                     Eight is recommended.
    //            epoch  - Time of required data values.
    //            aCoeff - Generated Lagrangian coefficients.
    // Return Value: True if generation of coefficients is successful. False otherwise.
    // Note: False is returned if requested epoch is outside of the range of available predictions. 
    // or when the time series is corrupt with overlapping data points.
    void generateLagrangianCoefficients( long index, long segment, YourTimeDateClass epoch);

    //  This function provides an interpolated value of X, corresponding to a given
    // time as defined by and incorporated into a set of Lagrangian coefficients. 
    // Arguments: aX -  array of data to be used associated with the required interpolated quantity.  
    //        aCoeff -  array of Lagragian coefficients incorporating the required time or epoch.
    // Return Value: The interpolated value of the quantity associated with array aX.
    double lagrangianInterpolation( const std::vector<double>& aX );
    YourTimeIntervalClass lagrangianInterpolation(const std::vector<YourTimeIntervalClass>&aX);

    //  Method to find the "best" index of an orbit point. It may
    // use the index previously used during a sequential process. The segment
    // associated with that index is also returned. 
    long getIndex( YourTimeDateClass epoch, long *lastIndex, long *segment );

    //  Method to find the epoch associated with an index of an orbit point.
    // The segment associated with that index is also returned. 
    YourTimeDateClass getEpoch( long index, long *segment = NULL );

    //  Routine to ensure azimuth and elevation angles are set to the 
    // best quadrant and are continuous throughout the data set.
    void normalizeAngles(double *previousAz, double *azimuth, double *elevation );

    //  Method to calculate position data given a site, and an index.
    // If no index is defined then the epoch is used. If no node data are available
    // (or index is out of range) then it returns false. 
    virtual bool processDataPoint(SiteHandle& site, YourTimeDateClass epoch, long index ){return true;};

    //  Method to obtain prediction data at a given epoch (using interpolation)
    // or index (copying from a node).
    bool getITRFData( SiteHandle& site, YourTimeDateClass epoch, long index );

    //  Method to interpolate data for all of the prediction data at the
    // given epoch. Returns true if successful. 
    virtual bool interpolateData(SiteHandle& site, YourTimeDateClass epoch ){return true;};

    //  Method to extract (copy) all of the prediction data at a
    // given node as defined by index. Returns true if successful. 
    virtual bool extractNodeData(SiteHandle& site, long index ){return true;}; 

    //  Method to create 'placeholder' data at each node, i.e.
    // at a time interval spacing given by step Size. 
    bool generateDummyNodes( SiteHandle& site );

    //  Method to update target characteristics such as proximity to the
    // sun and moon, illumination and visual magnitude.
    void updateTargetCharacteristics( SiteHandle& site );

    //  Methods to convert a matrix of rectangular position ( x,y,z)
    // and velocity (u,v,w) to spherical (range, azimuth, elevation) position
    // and velocity coordinates, and vice versa.
    void convertToSpherical(   Matrix<double>& pos, Matrix<double>& vel );
    void convertToRectangular( Matrix<double>& pos, Matrix<double>& vel );

    //  Method to determine the relativisitic scale (or ratio).
    double relativisticScale( YourTimeDateClass epoch, double longitude, double latitude );

    double m_sunHalo;
    double m_moonHalo;
    Type   m_trackType;
    bool   m_bestPrecision;
    bool   m_ignoreHorizon;
    bool   m_lightTimeCorrected;  // indicates if data corrected for light speed aberration.
    long   m_maxIndex;
    long   m_numInterp;    // number of nodes used in the interpolator
    long   m_numInterpMax; 
    unsigned long   m_numOfNodes;
    std::vector<double> m_lagCoeffs;
    std::vector<long>   m_lagIndex;

    YourTimeDateClass     m_periodStart;     // start time of requested predictions
    YourTimeDateClass     m_periodStop;      // stop time of requested predictions
    YourTimeDateClass     m_startPredicts;   // time of first prediction point
    YourTimeDateClass     m_stopPredicts;    // time of last prediction point
    YourTimeDateClass     m_startEpoch;      // time of first data point
    YourTimeDateClass     m_stopEpoch;       // time of last data point
    YourTimeIntervalClass m_stepSize;

    std::vector<YourTimeIntervalClass> m_segmentStartTimeBiasPreset;
    std::vector<YourTimeIntervalClass> m_segmentStopTimeBiasPreset;
    std::vector<YourTimeIntervalClass> m_segmentTimeBiasLocal;
    std::vector<YourTimeDateClass>     m_segmentStartTime;
    std::vector<YourTimeDateClass>     m_segmentStopTime;
    std::vector<double>                 m_segmentPosBiasAzim;
    std::vector<double>                 m_segmentPosBiasElev;
    std::vector<long>                   m_segmentStartIndex;
    std::vector<long>                   m_segmentStopIndex;


    std::vector<double> m_aX, m_aY, m_aZ;
    std::vector<double> m_aU, m_aV, m_aW;
    std::vector<double> m_aXPolar, m_aYPolar;
    std::vector<YourTimeIntervalClass> m_aDUTC;

};	// class Track

//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// The OrbitalTrack subclass is primarily a container of prediction data for
// targets orbiting the earth, including geostationary targets, and provides a 
// base class for lunar and transponder target types.
// The class supports both outbound and inbound legs if required and will support
// stellar aberration corrections.
//
// Remarks:
// The epoch of each data point is not stored, and must be determined using
// starting epoch and integer number of intervals. The prediction data is
// generated in segments (one for each prediction element (IRV, TLE etc) used.
// These segments will generally overlap in time hence the methods in the subclass
// must determine which segment is appropriate.
//-----------------------------------------------------------------------------
class OrbitalTrack : public Track
{

public:

    //  Default constructor 
    OrbitalTrack(){};

    //  Constructor
    OrbitalTrack( YourTimeDateClass     startPredictions,
                  YourTimeDateClass     stopPredictions,
                  YourTimeIntervalClass stepSize );

    //  Destructor
    virtual ~OrbitalTrack();

    //  Method to append a prediction node which includes
    // position, velocity and stella aberration correction data. 
    // Arguments: posVecCor -  a matrix containing the following data;
    // geocentric ITRF x,y,z positions at next time interval (Megameters)
    // geocentric ITRF u,v,w velocities at next time interval (Megameter/day)
    // x,y,z stellar abberration corrections at next time interval (m).
    // polar motion data, xp, yp (arcsecs) and UT1-UTC at next time interval.
    // biDirectional flag to indicate if independent inbound data is available. 
    // newSegment - flag indicating that this point is first in a new segment, i.e.
    //              predictions using a new orbital element.
    virtual void appendNode( YourTimeDateClass epoch,
                             bool newSegment, bool biDirectional, bool lightTimeCorrected,
                             const YourMatrix<double>& posVelCor, double xPolar,  double yPolar,
                             YourTimeIntervalClass ut1_utc );
    
protected:

    //  Method to transform ITRF geocentric coordinates
    // to topocentric coordinates (azimuth, elevation, range) wrt to a given site.
    bool convertITRFToTopocentric(SiteHandle& site, YourTimeDateClass epoch, long index);

    //  Method to create data at each node, i.e. at a time interval
    // spacing given by step Size. 
    virtual bool generateNodeData( SiteHandle& site ){return true;}

    //  Method to interpolate data for all of the prediction data at the
    // given epoch. Returns true if successful. 
    virtual bool interpolateData(SiteHandle& site, YourTimeDateClass epoch ); 

    //  Method to extract (copy) all of the prediction data at a
    // given node as defined by index. Returns true if successful. 
    virtual bool extractNodeData(SiteHandle& site, long index ); 

    //  Method to calculate position data given a site, epoch or index.
    // If index is undefined (set to a negtaive value) then the epoch is used.
    // If no node data are available (or index is out of range) then it returns false. 
    virtual bool processDataPoint(SiteHandle& site, YourTimeDateClass epoch, long index );

    //  Methods to calculate range data using target offset data.
    virtual void checkOffsets(bool inBound, SiteHandle& site, YourMatrix<double>& pos,
                              YourMatrix<double>& vel, YourTimeIntervalClass& range ){return;}; 

    //  Method to calculate time of flight of a light pulse between an earth bound
    // site and target. Note that a flag indicates if the light is inbound or outbound.
    bool getFlightTime(bool inBound, SiteHandle& site, YourTimeDateClass epoch, long index,
                       YourMatrix<double>& targetPos,     YourMatrix<double>& targetVel,
                       const YourMatrix<double>& sitePos, const YourMatrix<double>& siteVel,
                       YourTimeIntervalClass& range, double* delta = NULL );

    //  Method to calculate pointing angles (azimuth, elevation) and
    // pointing angle rates.
    void calculatePointingAngles(bool inBound, SiteHandle& site,
                                 YourMatrix<double>& targetPos,
                                 YourMatrix<double>& targetVel );

    double m_centreOfMassCorr;
    YourTimeIntervalClass m_targetResponseDelay;
    std::vector<bool>   m_segmentBiDirectional;

    std::vector<double> m_arX;
    std::vector<double> m_arY;
    std::vector<double> m_arZ;
    std::vector<double> m_arU;
    std::vector<double> m_arV;
    std::vector<double> m_arW;
    std::vector<YourTimeIntervalClass> m_aRelCorTx;
    std::vector<YourTimeIntervalClass> m_aRelCorRx;
    std::vector<double> m_aXAberTx;
    std::vector<double> m_aYAberTx;
    std::vector<double> m_aZAberTx;
    std::vector<double> m_aXAberRx;
    std::vector<double> m_aYAberRx;
    std::vector<double> m_aZAberRx;

};     // class OrbitalTrack

//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// The LunarTrack subclass supports Lunar Targets and Features, 
// and contains data additional to that provided in OrbitalTrack to support 
// Lunar feature offset pointing. Hence libration and rotation angles are also
// supported.
//-----------------------------------------------------------------------------
class LunarTrack : public OrbitalTrack
{
public:

    //  Default constructor
    LunarTrack(){};

    //  Constructor
    LunarTrack( YourTimeDateClass     startPredictions,
                YourTimeDateClass     stopPredictions,
                YourTimeIntervalClass stepSize );

    //  Destructor
    virtual ~LunarTrack();

    //  Method to append a prediction node which includes
    // position, velocity and stella aberration correction data. 
    // Arguments: posVecCor -  a matrix containing the following data;
    // geocentric ITRF x,y,z positions at next time interval (Megameters)
    // geocentric ITRF u,v,w velocities at next time interval (Megameter/day)
    // x,y,z stellar abberration corrections at next time interval (m).
    // polar motion data, xp, yp (arcsecs) and UT1-UTC at next time interval.
    // biDirectional flag to indicate if independent inbound data is available. 
    // gast - Greenwich Apparatent Sidereal Time at the next time interval.
    // targetOffsets - includes lunar target outbound and inbound offsets (m)
    //                 and rotation angles (degrees)
    // newSegment - flag indicating that this point is first in a new segment, i.e.
    //              predictions using a new orbital element.
    virtual void appendNode( YourTimeDateClass epoch,
                             bool newSegment, bool biDirectional, bool lightTimeCorrected,
                             const YourMatrix<double>& posVelCor, double xPolar,  double yPolar,
                             YourTimeIntervalClass ut1_utc,
                             YourTimeIntervalClass gast,
                             const YourMatrix<double>& targetOffsets );

protected:


    //  Methods to calculate range data using target offset data.
    virtual void checkOffsets(bool inBound, SiteHandle& site, YourMatrix<double>& pos,
                              YourMatrix<double>& vel, YourTimeIntervalClass& range );

    //  Method to interpolate data for all of the prediction data at the
    // given epoch. Returns true if successful. 
    virtual bool interpolateData(SiteHandle& site, YourTimeDateClass epoch );

    //  Method to extract (copy) all of the prediction data at a
    // given node as defined by index. Returns true if successful. 
    virtual bool extractNodeData(SiteHandle& site, long index ); 

    //  Method to generate the libration matrix.
    YourMatrix<double> librationMatrix( SiteHandle& site );
    
    //  Method to calculate position data given a site, epoch or index.
    // If index is undefined (set to a negtaive value) then the epoch is used.
    // If no node data are available (or index is out of range) then it returns false. 
    virtual bool processDataPoint(SiteHandle& site, YourTimeDateClass epoch, long index ); 


    std::vector<double> m_aXOffsetTx;
    std::vector<double> m_aYOffsetTx;
    std::vector<double> m_aZOffsetTx;
    std::vector<double> m_aXOffsetRx;
    std::vector<double> m_aYOffsetRx;
    std::vector<double> m_aZOffsetRx;
    std::vector<double> m_aRotAngle1;
    std::vector<double> m_aRotAngle2;
    std::vector<double> m_aRotAngle3;
    std::vector<YourTimeIntervalClass> m_aGast;

    std::vector<long>   m_rotationType;

};    // class LunarTrack

//-----------------------------------------------------------------------------
//  Class Definition
//-----------------------------------------------------------------------------
// The TransponderTrack subclass supports Transponder Targets, 
// and contains data additional to that provided in OrbitalTrack.
//-----------------------------------------------------------------------------
class TransponderTrack : public OrbitalTrack
{
public:

    //  Default constructor
    TransponderTrack(){};

    //  Constructor
    TransponderTrack( YourTimeDateClass     startPredictions,
                      YourTimeDateClass     stopPredictions,
                      YourTimeIntervalClass stepSize );

    //  Destructor
    virtual ~TransponderTrack();

    //  Method to append a prediction node which includes
    // position, velocity and stella aberration correction data. 
    // Arguments: posVecCor -  a matrix containing the following data;
    // geocentric ITRF x,y,z positions at next time interval (Megameters)
    // geocentric ITRF u,v,w velocities at next time interval (Megameter/day)
    // x,y,z stellar abberration corrections at next time interval (m).
    // polar motion data, xp, yp (arcsecs) and UT1-UTC at next time interval.
    // biDirectional flag to indicate if independent inbound data is available. 
    // transponder - includes transponder specific data at the next time interval.
    // newSegment - flag indicating that this point is first in a new segment, i.e.
    //              predictions using a new orbital element.
    virtual void appendNode( YourTimeDateClass epoch,
                             bool newSegment, bool biDirectional, bool lightTimeCorrected,
                             const YourMatrix<double>& posVelCor, double xPolar,  double yPolar,
                             YourTimeIntervalClass ut1_utc,
                             const std::vector<double>& transponder );

protected:

    //  Method to interpolate data for all of the prediction data at the
    // given epoch. Returns true if successful. 
    virtual bool interpolateData(SiteHandle& site, YourTimeDateClass epoch ); 

    //  Method to extract (copy) all of the prediction data at a
    // given node as defined by index. Returns true if successful. 
    virtual bool extractNodeData(SiteHandle& site, long index ); 

    //  Method to calculate position data given a site, epoch or index.
    // If index is undefined (set to a negtaive value) then the epoch is used.
    // If no node data are available (or index is out of range) then it returns false. 
    virtual bool processDataPoint(SiteHandle& site, YourTimeDateClass epoch, long index ); 

    std::vector<double> m_aTransponderOscillatorRelCor;
    std::vector<double> m_transponderPulseFrequency;
    std::vector<double> m_transponderUtcOffset;
    std::vector<double> m_transponderTransmitDelay;
    std::vector<double> m_transponderOscillatorDrift;

};    // class TransponderTrack

#endif // _TRACK_H_
