#include <stdio.h>
#include <math.h>
/* ========================================================================
CC  PURPOSE   : INTERPOLATE SATELLITE POSITION, DETERMINE PREDICTION
CC              ELEMENTS
CC
CC  PARAMETERS:
CC         IN : TT       : EPOCH (TRANSMIT=FIRING TIME)            R*8
CC              IUNITS   : UNITS OF TIME (0 = SECONDS, #0 = DAYS)  I*4
CC                         (FOR TT AND TTAB!)
CC              MODE     : 0: NO FLIGHT TIME APPLIED               I*4
CC                            FLIGHT TIME = 2 * INSTANT RANGE
CC                                          AT EPOCH TT
CC                         1: OUTBOUND VECTOR (AZIMUTH, ELEVATION)
CC                            FLIGHT TIME = 2 * INSTANT RANGE AT
CC                                          BOUNCE TIME
CC                         2: OUTBOUND AND INBOUND VECTOR
CC                            FLIGHT TIME = OUTBOUND + INBOUND
CC              STAGEO   : GEOCENTRIC STATION COORDINATES XYZ (M)  R*8(*)
CC              STALON   : STATION LONGITUDE (EAST  > 0, RADIANS)  R*8
CC              STALAT   : STATION LATITUDE  (NORTH > 0, RADIANS)  R*8
CC              COFM_CORR: RADIAL CENTER OF MASS CORRECTION (M)    R*8
CC              TTAB     : TABULATED EPOCHS (SECONDS OR DAYS,      R*8
CC                                           SEE IUNITS)
CC              OBJTAB_X : TABULATED X-COORD. OF OBJECT (M)        R*8(*)
CC              OBJTAB_Y : TABULATED Y-COORD. OF OBJECT (M)        R*8(*)
CC              OBJTAB_Z : TABULATED Z-COORD. OF OBJECT (M)        R*8(*)
CC              NTAB     : NUMBER OF TABULATED EPOCHS/COORDINATES  I*4
CC              NINT     : NUMBER OF TABULATED VALUES TO USE       I*4
CC                         (ARGUMENT FOR LAGRANGE INTERPOLATION)
CC        OUT : RANGE    : ONE-WAY RANGE (M)                       R*8
CC              FLTIME   : TWO-WAY FLIGHT TIME (SEC)               R*8
CC              AZIOUT   : OUTBOUND AZIMUTH   (DEG)                R*8
CC              ELEOUT   : OUTBOUND ELEVATION (DEG)                R*8
CC              DIFAZI   : POINT-BEHIND (AZIMUTH,   DEGREES)       R*8
CC              DIFELE   : POINT-BEHIND (ELEVATION, DEGREES)       R*8
CC                         DIFFERENCE INBOUND MINUS OUTBOUND AT
CC                         EPOCH TT
CC              IRCODE   : RETURN CODE 0 = OK                      I*4
CC                                     1 = INTERPOLATION NOT IN
CC                                         CENTER INTERVAL
CC                                     2 = ERROR IN INTERPOLATION
CC
CC  CALLS:      HERMITE, DDREH, DMLMAV
CC
CC  AUTHOR:     W. GURTNER
CC              ASTRONOMICAL INSTITUTE
CC              UNIVERSITY OF BERNE
CC
CC  CREATED:    25-JUL-2005
CC
CC  MODIFIED:    2-AUG-2005 : TIME UNITS (SEC OR DAYS), MODE
CC              17-Oct-2005 : Converted to c; R. Ricklefs UT/CSR
CC              29-NOV-2005 : RADIAL CENTER OF MASS CORRECTION
CC		30-JUN-2009 : Comments: DIFAZI and DIFELE are degrees not asec.
=========================================================================*/

void
cpf_inter(double tt, int iunits, int mode, double stageo[3],
          double stalat, double stalon, double cofm_corr,
          double *ttab, double *objtab_x,double *objtab_y,double *objtab_z,
          int ntab,int nint,
          double *range, double *fltime, double *aziout, double *eleout,
          double *difazi, double *difele,
          int *ircode)
{
      double objbnc[3],objtrt[3],topinb[3],topout[3],toptrt[3];
      double r3sd[3][3],r3[3][3],starot[3];
      double x, z, vx, vy, vz, disttrt, eletrt, azitrt, tout, tb;
      double distout, dsidt, tinb, tr, distinb, distot;
      double c= 299792458.e0;	/* Speed of light */
      int i, ircx, ircy, ircz;
      static int ifirst= 1;
      static double pi, prvlon, prvlat; 
      static double r2la[3][3], r3lo[3][3], r3pi[3][3];

      if (ifirst == 1) {

        pi=4.e0*atan(1.e0);
        prvlon=999.e0;
        prvlat=999.e0;
        ifirst=0;
      }
/*
 *  DDREH:  Computation of rotation matrices
 *  r3lo around longitude, r2la around pi/2-latitude, r3pi around pi */
      if(stalon != prvlon || stalat != prvlat) {
        ddreh(3,stalon,r3lo);
        ddreh(2,pi/2-stalat,r2la);
        ddreh(3,pi,r3pi);
        prvlon=stalon;
        prvlat=stalat;
      }
/*
 *  INSTANT VECTOR AT TRANSMIT TIME
 *  *******************************
 */
/*  Interpolate geocentric position of the object for transmit time */
      hermite(1,ttab,objtab_x,&z,ntab,nint,tt,&objtrt[0],&vx,&ircx);
      hermite(1,ttab,objtab_y,&z,ntab,nint,tt,&objtrt[1],&vy,&ircy);
      hermite(1,ttab,objtab_z,&z,ntab,nint,tt,&objtrt[2],&vz,&ircz);
      if (ircx+ircy+ircz != 0) {
        printf(" SR CPF_INTER: Return codes from SR HERMITE: %3i %3i %3i\n",
          ircx,ircy,ircz);
        *ircode=2;
        return;
      }

/*  Topocentric vector station/object both at transmit time */
      toptrt[0]=objtrt[0]-stageo[0];
      toptrt[1]=objtrt[1]-stageo[1];
      toptrt[2]=objtrt[2]-stageo[2];

/*   Instant distance from station to object at transmit time */
      disttrt=sqrt(toptrt[0]*toptrt[0]+toptrt[1]*toptrt[1]+toptrt[2]*toptrt[2]);
/*
 *  Topocentric vector in local systeM
 */
/*  dmlmav: multiply rotation matrix with vector */
      dmlmav(toptrt,r3lo,toptrt);
      dmlmav(toptrt,r2la,toptrt);
      dmlmav(toptrt,r3pi,toptrt);

/*  azimuth and elevation (degrees) */
      eletrt=atan(toptrt[2]/sqrt(toptrt[0]*toptrt[0]+toptrt[1]*toptrt[1]))
	     *180/pi;
      azitrt=atan2(-toptrt[1],toptrt[0])*180/pi;
      if(azitrt < 0.e0) azitrt+=360;

/*  Instant vector only */
      if (mode == 0) {
/*  One-way range */
        *range= disttrt- cofm_corr;
/*  Round trip flight time (sec) */
        *fltime= 2* (*range)/c;
/*  Direction */
        *aziout= azitrt;
        *eleout= eletrt;
/*  no point behind */
        *difazi= 0;
        *difele= 0;
        return;
      }
/*
 *  Outbound vector
 *  ***************
 */
/*  Initialize */
      starot[0]=stageo[0];
      starot[1]=stageo[1];
      starot[2]=stageo[2];
      tout=disttrt/c;
/*
 *  Iteration
 */
      for (i=0; i<2; i++) {

/*  BOUNCE TIME */
        if (iunits == 0) {
          tb= tt+tout;
        } else {
          tb= tt+tout/86400.e0;
        }

/*  Interpolate geocentric position of the object for bounce time tb */
        hermite(1,ttab,objtab_x,&z,ntab,nint,tb,&objbnc[0],&vx,&ircx);
        hermite(1,ttab,objtab_y,&z,ntab,nint,tb,&objbnc[1],&vy,&ircy);
        hermite(1,ttab,objtab_z,&z,ntab,nint,tb,&objbnc[2],&vz,&ircz);
        if (ircx+ircy+ircz != 0) {
          printf(" SR CPF_INTER: Return codes from SR HERMITE: %3i %3i %3i\n",
            ircx,ircy,ircz);
          *ircode= 2;
          return;
        }
       
/*  Topocentric outbound vector */
        topout[0]=objbnc[0]-starot[0];
        topout[1]=objbnc[1]-starot[1];
        topout[2]=objbnc[2]-starot[2];

/*  Distance from station (tt) to object (tb) */
        distout=sqrt(topout[0]*topout[0]+topout[1]*topout[1]+topout[2]*topout[2]);

/*  Outbound flight time (sec) */
        tout= distout/c;

/*  rotate station during flight time (radians) */
        dsidt=6.300388e0*(tout/86400.e0);
        ddreh(3,dsidt,r3sd);
        dmlmav(stageo,r3sd,starot);

      }
/*
 *  Topocentric outbound vector in local system
 */
      dmlmav(topout,r3lo,topout);
      dmlmav(topout,r2la,topout);
      dmlmav(topout,r3pi,topout);

/*  Outbound azimuth and elevation (laser beam pointing direction)
 *  **************************************************************
 */
      *eleout=atan(topout[2]/sqrt(topout[0]*topout[0]+topout[1]*topout[1]))
             *180/pi;
      *aziout=atan2(-topout[1],topout[0])*180/pi;
      if(*aziout < 0.e0) *aziout+=360;
/*      printf (" Azi/Ele outbound vector : %10.4f %10.4f\n", aziout, eleout);
 */
/*
 *  Total point ahead / behind
 *  ************************** 
 *  DIFFERENCE BETWEEN RECEIVE AND TRANSMIT DIRECTION AT TRANSMIT TIME
 */
      *difazi= 2*(azitrt-*aziout);
      if (*difazi < -360) *difazi+=720;
      if (*difazi > +360) *difazi-=720;
      *difele= 2*(eletrt-*eleout);
/*      printf (" Point-behind (AZI,ELE arcsec) : %8.1f %8.1f\n", 
 *        difazi*3600, difele*3600); */

      if(mode == 1) {
/*
 *  AVERAGE DISTANCE FROM STATION TO OBJECT
 *  (BOTH AT BOUNCE TIME: GOOD APPROXIMATION)
 *  *****************************************
 */
        *range= sqrt((objbnc[0]-stageo[0])*(objbnc[0]-stageo[0])+
                     (objbnc[1]-stageo[1])*(objbnc[1]-stageo[1])+
                     (objbnc[2]-stageo[2])*(objbnc[2]-stageo[2]));
/*
 *  Radial center of mass correction
 *  ******************************** */
        /*printf("range %15.1f cofm %7.4f\n",*range,cofm_corr);*/
        (*range)-= cofm_corr;
/*
 *  Round trip flight time (sec)
 *  **************************** */
        *fltime=2* (*range)/c;
        /*printf("range %15.1f ft %15.10f\n",*range,*fltime);*/

        return;

      }

      if (mode == 2) {
/*
 *  Inbound vector: station at receiving time
 *  *****************************************
 */
        tinb= tout;

/*  Iteration */
        for (i=0; i<2; i++) {

/*  Receive time */
          if (iunits == 0) {
            tr=tb+tinb;
          } else {
            tr=tb+tinb/86400.e0;
          }

/*  Rotate station during inbound flight time (other direction) */
          dsidt=6.300388e0*(tinb/86400.e0);
          ddreh(3,-dsidt,r3);
          dmlmav(stageo,r3,starot);

/*  Topocentric inbound vector in geocentric system */
          topinb[0]=objbnc[0]-starot[0];
          topinb[1]=objbnc[1]-starot[1];
          topinb[2]=objbnc[2]-starot[2];

/*  Distance from object to station */
          distinb=sqrt(topinb[0]*topinb[0]+topinb[1]*topinb[1]
                    +topinb[2]*topinb[2]);

/*  Inbound flight time (sec) */
          tinb= distinb/c;

       }
/*
 *  Total distance from station to object to station (correct computation)
 */
        distot=distout+distinb;
        *fltime=(distot- 2.e0*cofm_corr)/c;
        *range =distot/2.e0- cofm_corr;
      }

}
